@@ -1,8 +1,8 @@
 import pandas as pd, numpy as np
 import multiprocessing, os, warnings
-from . import cy
+from . import cy_double, cy_float, _check_openmp
 import ctypes, types, inspect
-from hpfrec import HPF, cython_loops_float as cython_loops
+from hpfrec import HPF, cython_loops_float, cython_loops_double
 pd.options.mode.chained_assignment = None
 
 
@@ -137,6 +137,11 @@ class CTPF:
 		Calculate log-likelihood every N iterations.
 	verbose : bool
 		Whether to print convergence messages.
+	use_float : bool
+		Whether to use the C float type (typically ``np.float32``). Using float types (as compared to double)
+		results in less memory usage and faster operations, but it has less numeric precision and the results
+		will be slightly worse compared to using double.
+		If passing ``False``, will use C double (typically ``np.float64``).
 	random_seed : int or None
 		Random seed to use when starting the parameters.
 	ncores : int
@@ -245,7 +250,7 @@ class CTPF:
 	def __init__(self, k=50, a=.3, b=.3, c=.3, d=.3, e=.3, f=.3, g=.3, h=.3,
 				 stop_crit='train-llk', stop_thr=1e-3, reindex=True,
 				 miniter=25, maxiter=70, check_every=10, verbose=True,
-				 random_seed=None, ncores=-1, initialize_hpf=True,
+				 use_float=True, random_seed=None, ncores=-1, initialize_hpf=True,
 				 standardize_items=False, rescale_factors=False,
 				 missing_items='include', step_size=lambda x: 1-1/np.sqrt(x+1),
 				 allow_inconsistent_math=False, full_llk=False,
@@ -297,6 +302,13 @@ class CTPF:
 		assert ncores>0
 		assert isinstance(ncores, int)
 
+		if (ncores > 1) and not (_check_openmp.get()):
+			msg_omp  = "Attempting to use more than 1 thread, but "
+			msg_omp += "package was built without multi-threading "
+			msg_omp += "support - see the project's GitHub page for "
+			msg_omp += "more information."
+			warnings.warn(msg_omp)
+
 		if random_seed is not None:
 			assert isinstance(random_seed, int)
 
@@ -371,6 +383,7 @@ class CTPF:
 		self.stop_thr = stop_thr
 		self.save_folder = save_folder
 		self.verbose = verbose
+		self.use_float = bool(use_float)
 		self.produce_dicts = bool(produce_dicts)
 		self.full_llk = bool(full_llk)
 		self.keep_all_objs = bool(keep_all_objs)
@@ -701,10 +714,11 @@ class CTPF:
 		return col1, col2, subj
 
 	def _cast_df(self, df, ttl):
+		cy = cy_float if self.use_float else cy_double
 		col1, col2, subj = self._cols_from_ttl(ttl)
 		df[col1] = df[col1].values.astype(cy.obj_ind_type)
 		df[col2] = df[col2].values.astype(cy.obj_ind_type)
-		df['Count'] = df.Count.astype(ctypes.c_float)
+		df['Count'] = df.Count.astype(ctypes.c_float if self.use_float else ctypes.c_double if self.use_float else ctypes.c_double)
 		return df
 
 	def _check_df(self, df, ttl):
@@ -776,9 +790,9 @@ class CTPF:
 			else:
 				self._unexpected_err_msg()
 
-			df[col2] = pd.Categorical(df[col2].values, curr_mapping2).codes
+			df[col2] = pd.Categorical(df[col2], curr_mapping2).codes
 			new_ids2 = df[col2].values == -1
-			if new_ids2.sum() > 0:
+			if new_ids2.any():
 				df = df.loc[~new_ids2].reset_index(drop=True)
 				if df.shape[0] > 0:
 					msg = "'" + ttl + "' has " + subj2 + "s that were not present in the training data."
@@ -789,9 +803,9 @@ class CTPF:
 
 			new_ids1 = df[col1].unique()
 			repeated = np.in1d(new_ids1, curr_mapping1)
-			if repeated.sum() > 0:
+			if repeated.any():
 				repeated = new_ids1[repeated]
-				df = df.loc[~np.in1d(df[col1].values, repeated)].reset_index(drop=True)
+				df = df.loc[~np.in1d(df[col1].to_numpy(), repeated)].reset_index(drop=True)
 				if df.shape[0] > 0:
 					msg = "'" + ttl + "' contains " + subj1 + "s that were already present in the training set."
 					msg += " These will be ignored."
@@ -804,9 +818,9 @@ class CTPF:
 				ttl2 = "counts_df"
 				df2 = self._check_df(df2, ttl=ttl2)
 
-				df2['ItemId'] = pd.Categorical(df2[col1].values, self.item_mapping_).codes
+				df2['ItemId'] = pd.Categorical(df2[col1], self.item_mapping_).codes
 				invalid_items = df2.ItemId == -1
-				if invalid_items.sum() > 0:
+				if invalid_items.any() > 0:
 					df2 = df2.loc[~invalid_items].reset_index(drop=True)
 					if df2.shape[0] > 0:
 						msg = "'" + ttl2 + "' has " + "item" + "s that were not present in the training data."
@@ -817,7 +831,7 @@ class CTPF:
 
 				new_ids11 = df2[col1].unique()
 				repeated = np.in1d(new_ids11, curr_mapping1)
-				if repeated.sum() > 0:
+				if repeated.any() > 0:
 					repeated = new_ids1[repeated]
 					df2 = df2.loc[~np.in1d(df2[col1].values, repeated)].reset_index(drop=True)
 					if df2.shape[0] > 0:
@@ -828,13 +842,15 @@ class CTPF:
 						raise ValueError("'" + ttl2 + "' doesn't contain any new " + subj1 + "s.")
 
 				new_ids1 = np.unique(np.r_[new_ids1, new_ids11])
+				new_ids1 = np.setdiff1d(new_ids1, curr_mapping1)
 				new_mapping = np.r_[curr_mapping1, new_ids1]
-				df[col1] = pd.Categorical(df[col1].values, new_mapping).codes
-				df2[col1] = pd.Categorical(df2[col1].values, new_mapping).codes
+				df[col1] = pd.Categorical(df[col1], new_mapping).codes
+				df2[col1] = pd.Categorical(df2[col1], new_mapping).codes
 				df2 = self._cast_df(df2, ttl=ttl2)
 			else:
+				new_ids1 = np.setdiff1d(new_ids1, curr_mapping1)
 				new_mapping = np.r_[curr_mapping1, new_ids1]
-				df[col1] = pd.Categorical(df[col1].values, new_mapping).codes
+				df[col1] = pd.Categorical(df[col1], new_mapping).codes
 		
 		else:
 			new_mapping = None
@@ -852,10 +868,11 @@ class CTPF:
 		self.seen = self._counts_df[['UserId', 'ItemId']].copy()
 		self.seen.sort_values(['UserId', 'ItemId'], inplace=True)
 		self.seen.reset_index(drop = True, inplace = True)
-		self._n_seen_by_user = self.seen.groupby('UserId')['ItemId'].agg(lambda x: len(x)).values
+		self._n_seen_by_user = self.seen.groupby('UserId')['ItemId'].agg(lambda x: len(x)).values.astype(int)
 		self._st_ix_user = np.cumsum(self._n_seen_by_user)
 		self._st_ix_user = np.r_[[0], self._st_ix_user[:self._st_ix_user.shape[0]-1]]
-		self.seen = self.seen.ItemId.values
+		self._st_ix_user = self._st_ix_user.astype(int)
+		self.seen = self.seen.ItemId.values.astype(int)
 		return None
 
 	def _exclude_missing_from_index(self):
@@ -877,16 +894,17 @@ class CTPF:
 
 	def _initalize_parameters(self):
 		## TODO: make this function more modular
+		## TODO: improve the way of adding random noise in the initialization
+		## Note: the initialization here corresponds to the one used in the original HPF code
 
-		rng = np.random.default_rng(seed = self.random_seed \
-									if (self.random_seed is not None) and (self.random_seed > 0) else None)
+		rng = np.random.Generator(np.random.MT19937(seed = self.random_seed))
 
 		if self.initialize_hpf:
 			if self.verbose:
 				print("Initializing Theta and Beta through HPF...")
 				print("")
 			h = HPF(k=self.k, verbose=self.verbose, reindex=True, produce_dicts=False, stop_crit='diff-norm',
-					stop_thr=self.stop_thr, random_seed=self.random_seed, keep_all_objs=False,
+					stop_thr=self.stop_thr, random_seed=self.random_seed, keep_all_objs=False, use_float=self.use_float,
 					sum_exp_trick=self.sum_exp_trick, allow_inconsistent_math=self.allow_inconsistent_math,)
 			h.fit(self._words_df.rename(columns={'ItemId':'UserId', 'WordId':'ItemId'}).copy())
 			if (h.nusers == self.nitems) and (h.nitems == self.nwords):
@@ -913,35 +931,48 @@ class CTPF:
 					words_include = self._words_df.WordId.loc[np.in1d(self._words_df.ItemId, items_words_df[items_intersect])].unique()
 					words_take = pd.Categorical(words_include, h.item_mapping_).codes
 
-				self.Theta_shp = (self.c * 2*rng.beta(20, 20, size=(self.nitems, self.k))).astype(ctypes.c_float)
-				self.Theta_shp[h.user_mapping_[items_take],:] = h.Theta[items_take].copy()
+				self.Theta_shp = self.c + rng.uniform(0, 0.01, size=(self.nitems, self.k))
+				if self.use_float:
+					self.Theta_shp = self.Theta_shp.astype(ctypes.c_float)
+				self.Theta_shp[h.user_mapping_[items_take],:] = h.Theta[items_take]
 
-				self.Beta_shp = (self.a * 2*rng.beta(20, 20, size=(self.nwords, self.k))).astype(ctypes.c_float)
-				self.Beta_shp[h.item_mapping_[words_take],:] = h.Beta[words_take].copy()
+				self.Beta_shp = self.a + rng.uniform(0, 0.01, size=(self.nwords, self.k))
+				if self.use_float:
+					self.Beta_shp = self.Beta_shp.astype(ctypes.c_float)
+				self.Beta_shp[h.item_mapping_[words_take],:] = h.Beta[words_take]
 				
 				self.Theta_rte = self.d + self.Beta_shp.sum(axis=0, keepdims=True)
 				self.Beta_rte = self.b + self.Theta_shp.sum(axis=0, keepdims=True)
 
 			if np.isnan(self.Theta_shp).sum().sum() > 0:
 				warnings.warn("NaNs produced in initialization of Theta, will use a random start.")
-				self.Theta_shp = (self.c * 2*rng.beta(20, 20, size=(self.nitems, self.k))).astype(ctypes.c_float)
-				self.Theta_rte = (self.d * 2*rng.beta(20, 20, size=(1, self.k))).astype(ctypes.c_float)
+				self.Theta_shp = self.c + rng.uniform(0, 0.01, size=(self.nitems, self.k))
+				self.Theta_rte = self.d + rng.uniform(0, 0.01, size=(1, self.k))
+				if self.use_float:
+					self.Theta_shp = self.Theta_shp.astype(ctypes.c_float)
+					self.Theta_rte = self.Theta_rte.astype(ctypes.c_float)
 			if np.isnan(self.Beta_shp).sum().sum() > 0:
 				warnings.warn("NaNs produced in initialization of Beta, will use a random start.")
-				self.Beta_shp = (self.a * 2*rng.beta(20, 20, size=(self.nwords, self.k))).astype(ctypes.c_float)
-				self.Beta_rte = (self.b * 2*rng.beta(20, 20, size=(1, self.k))).astype(ctypes.c_float)
+				self.Beta_shp = self.a + rng.uniform(0, 0.01, size=(self.nwords, self.k))
+				self.Beta_rte = self.b + rng.uniform(0, 0.01, size=(1, self.k))
+				if self.use_float:
+					self.Beta_shp = self.Beta_shp.astype(ctypes.c_float)
+					self.Beta_rte = self.Beta_rte.astype(ctypes.c_float)
 			if self.verbose:
 				print("**********************************")
 				print("")
 
 			if self._has_user_df:
-				self.Omega_shp = (self.e * 2*rng.beta(20, 20, size=(self.nusers, self.k))).astype(ctypes.c_float)
-				self.Omega_rte = (self.f * 2*rng.beta(20, 20, size=(1, self.k))).astype(ctypes.c_float)
+				self.Omega_shp = self.e + rng.uniform(0, 0.01, size=(self.nusers, self.k))
+				self.Omega_rte = self.f + rng.uniform(0, 0.01, size=(1, self.k))
+				if self.use_float:
+					self.Omega_shp = self.Omega_shp.astype(ctypes.c_float)
+					self.Omega_rte = self.Omega_rte.astype(ctypes.c_float)
 				if self.verbose:
 					print("Initializing Kappa through HPF...")
 					print("")
 				h = HPF(k=self.k, verbose=self.verbose, reindex=True, produce_dicts=False, stop_crit='diff-norm',
-					stop_thr=self.stop_thr, random_seed=self.random_seed, keep_all_objs=False,
+					stop_thr=self.stop_thr, random_seed=self.random_seed, keep_all_objs=False, use_float=self.use_float,
 					sum_exp_trick=self.sum_exp_trick, allow_inconsistent_math=self.allow_inconsistent_math)
 				h.fit(self._user_df.rename(columns={'AttributeId':'ItemId'}).copy())
 				
@@ -961,7 +992,9 @@ class CTPF:
 						attr_include = self._user_df.AttributeId.loc[np.in1d(self._user_df.UserId, users_user_df[users_intersect])].unique()
 						attr_take = pd.Categorical(attr_include, h.item_mapping_).codes
 
-					self.Kappa_shp = (self.a * 2*rng.beta(20, 20, size=(self.nuserattr, self.k))).astype(ctypes.c_float)
+					self.Kappa_shp = self.a + rng.uniform(0, 0.01, size=(self.nuserattr, self.k))
+					if self.use_float:
+						self.Kappa_shp = self.Kappa_shp.astype(ctypes.c_float)
 					self.Kappa_shp[h.item_mapping_[attr_take],:] = h.Beta[attr_take].copy()
 					self.Kappa_rte = self.b + h.Theta.sum(axis=0, keepdims=True)
 					del h
@@ -969,36 +1002,55 @@ class CTPF:
 
 				if np.isnan(self.Kappa_shp).sum().sum() > 0:
 					warnings.warn("NaNs produced in initialization of Kappa, will use a random start.")
-					self.Kappa_shp = (self.a * 2*rng.beta(20, 20, size=(self.nuserattr, self.k))).astype(ctypes.c_float)
-					self.Kappa_rte = (self.b * 2*rng.beta(20, 20, size=(1, self.k))).astype(ctypes.c_float)
+					self.Kappa_shp = self.a + rng.uniform(0, 0.01, size=(self.nuserattr, self.k))
+					self.Kappa_rte = self.b + rng.uniform(0, 0.01, size=(1, self.k))
+					if self.use_float:
+						self.Kappa_shp = self.Kappa_shp.astype(ctypes.c_float)
+						self.Kappa_rte = self.Kappa_rte.astype(ctypes.c_float)
 				if self.verbose:
 					print("**********************************")
 					print("")
 			else:
-				self.Kappa_shp = np.empty((0,0), dtype=ctypes.c_float)
-				self.Kappa_rte = np.empty((0,0), dtype=ctypes.c_float)
+				self.Kappa_shp = np.empty((0,0), dtype=ctypes.c_float if self.use_float else ctypes.c_double)
+				self.Kappa_rte = np.empty((0,0), dtype=ctypes.c_float if self.use_float else ctypes.c_double)
 		else:
-			self.Beta_shp = (self.a * 2*rng.beta(20, 20, size=(self.nwords, self.k))).astype(ctypes.c_float)
-			self.Theta_shp = (self.c * 2*rng.beta(20, 20, size=(self.nitems, self.k))).astype(ctypes.c_float)
-			self.Beta_rte = (self.b * 2*rng.beta(20, 20, size=(1, self.k))).astype(ctypes.c_float)
-			self.Theta_rte = (self.d * 2*rng.beta(20, 203, size=(1, self.k))).astype(ctypes.c_float)
+			self.Beta_shp = self.a + rng.uniform(0, 0.01, size=(self.nwords, self.k))
+			self.Theta_shp = self.c + rng.uniform(0, 0.01, size=(self.nitems, self.k))
+			self.Beta_rte = self.b + rng.uniform(0, 0.01, size=(1, self.k))
+			self.Theta_rte = self.d + rng.uniform(0, 0.01, size=(1, self.k))
+			if self.use_float:
+				self.Beta_shp = self.Beta_shp.astype(ctypes.c_float)
+				self.Theta_shp = self.Theta_shp.astype(ctypes.c_float)
+				self.Beta_rte = self.Beta_rte.astype(ctypes.c_float)
+				self.Theta_rte = self.Theta_rte.astype(ctypes.c_float)
 			if self._has_user_df:
-				self.Kappa_shp = (self.a * 2*rng.beta(20, 20, size=(self.nuserattr, self.k))).astype(ctypes.c_float)
-				self.Kappa_rte = (self.b * 2*rng.beta(20, 20, size=(1, self.k))).astype(ctypes.c_float)
+				self.Kappa_shp = self.a + rng.uniform(0, 0.01, size=(self.nuserattr, self.k))
+				self.Kappa_rte = self.b + rng.uniform(0, 0.01, size=(1, self.k))
+				if self.use_float:
+					self.Kappa_shp = self.Kappa_shp.astype(ctypes.c_float)
+					self.Kappa_rte = self.Kappa_rte.astype(ctypes.c_float)
 			else:
-				self.Kappa_shp = np.empty((0,0), dtype=ctypes.c_float)
-				self.Kappa_rte = np.empty((0,0), dtype=ctypes.c_float)
+				self.Kappa_shp = np.empty((0,0), dtype=ctypes.c_float if self.use_float else ctypes.c_double)
+				self.Kappa_rte = np.empty((0,0), dtype=ctypes.c_float if self.use_float else ctypes.c_double)
 		
-		self.Eta_shp = (self.e * 2*rng.beta(20, 20, size=(self.nusers, self.k))).astype(ctypes.c_float)
-		self.Epsilon_shp = (self.g * 2*rng.beta(20, 20, size=(self.nitems, self.k))).astype(ctypes.c_float)
-		self.Eta_rte = (self.f * 2*rng.beta(20, 20, size=(1, self.k))).astype(ctypes.c_float)
-		self.Epsilon_rte = (self.h * 2*rng.beta(20, 20, size=(1, self.k))).astype(ctypes.c_float)
+		self.Eta_shp = self.e + rng.uniform(0, 0.01, size=(self.nusers, self.k))
+		self.Epsilon_shp = self.g + rng.uniform(0, 0.01, size=(self.nitems, self.k))
+		self.Eta_rte = self.f + rng.uniform(0, 0.01, size=(1, self.k))
+		self.Epsilon_rte = self.h + rng.uniform(0, 0.01, size=(1, self.k))
+		if self.use_float:
+			self.Eta_shp = self.Eta_shp.astype(ctypes.c_float)
+			self.Epsilon_shp = self.Epsilon_shp.astype(ctypes.c_float)
+			self.Eta_rte = self.Eta_rte.astype(ctypes.c_float)
+			self.Epsilon_rte = self.Epsilon_rte.astype(ctypes.c_float)
 		if self._has_user_df:
-			self.Omega_shp = (self.e * 2*rng.beta(20, 20, size=(self.nusers, self.k))).astype(ctypes.c_float)
-			self.Omega_rte = (self.f * 2*rng.beta(20, 20, size=(1, self.k))).astype(ctypes.c_float)
+			self.Omega_shp = self.e + rng.uniform(0, 0.01, size=(self.nusers, self.k))
+			self.Omega_rte = self.f + rng.uniform(0, 0.01, size=(1, self.k))
+			if self.use_float:
+				self.Omega_shp = self.Omega_shp.astype(ctypes.c_float)
+				self.Omega_rte = self.Omega_rte.astype(ctypes.c_float)
 		else:
-			self.Omega_shp = np.empty((0,0), dtype=ctypes.c_float)
-			self.Omega_rte = np.empty((0,0), dtype=ctypes.c_float)
+			self.Omega_shp = np.empty((0,0), dtype=ctypes.c_float if self.use_float else ctypes.c_double)
+			self.Omega_rte = np.empty((0,0), dtype=ctypes.c_float if self.use_float else ctypes.c_double)
 
 		self._divide_parameters(add_beta=False)
 
@@ -1106,6 +1158,9 @@ class CTPF:
 		if has_coo:
 			self.reindex = False
 
+		cy = cy_float if self.use_float else cy_double
+		cython_loops = cython_loops_float if self.use_float else cython_loops_double
+
 		## running each sub-process
 		if self.verbose:
 			self._print_st_msg()
@@ -1118,11 +1173,11 @@ class CTPF:
 			self.val_set = pd.DataFrame({
 				'UserId': np.empty(0, dtype=cy.obj_ind_type),
 				'ItemId': np.empty(0, dtype=cy.obj_ind_type),
-				'Count': np.empty(0, dtype=ctypes.c_float)})
+				'Count': np.empty(0, dtype=ctypes.c_float if self.use_float else ctypes.c_double)})
 		if not self._has_user_df:
 			self._user_df = pd.DataFrame({'UserId':np.empty(0, dtype=cy.obj_ind_type),
 				'AttributeId':np.empty(0, dtype=cy.obj_ind_type),
-				'Count':np.empty(0, dtype=ctypes.c_float)})
+				'Count':np.empty(0, dtype=ctypes.c_float if self.use_float else ctypes.c_double)})
 		if self.verbose:
 			print("Initializing parameters...")
 		self._initalize_parameters()
@@ -1135,7 +1190,7 @@ class CTPF:
 			else:
 				self._user_df = pd.DataFrame({'UserId':np.empty(0, dtype=cy.obj_ind_type),
 											  'AttributeId':np.empty(0, dtype=cy.obj_ind_type),
-											  'Count':np.empty(0, dtype=ctypes.c_float)})
+											  'Count':np.empty(0, dtype=ctypes.c_float if self.use_float else ctypes.c_double)})
 
 		## fitting the model
 		self.niter = cy.fit_ctpf(
@@ -1182,7 +1237,7 @@ class CTPF:
 			if exclude_seen:
 				n_ext = int(np.min([n + self._n_seen_by_user[user], self._M2.shape[0]]))
 				rec = np.argpartition(allpreds, n_ext-1)[:n_ext]
-				seen = self.seen[int(self._st_ix_user[user]) : int(self._st_ix_user[user] + self._n_seen_by_user[user])]
+				seen = self.seen[self._st_ix_user[user] : self._st_ix_user[user] + self._n_seen_by_user[user]]
 				rec = np.setdiff1d(rec, seen)
 				rec = rec[np.argsort(allpreds[rec])[:n]]
 				if self.reindex:
@@ -1231,7 +1286,7 @@ class CTPF:
 				allpreds = - user_vec.dot(self._M2[items_pool].T)
 			n = np.min([n, items_pool.shape[0]])
 			if exclude_seen:
-				n_ext = np.min([n + self._n_seen_by_user[user], items_pool.shape[0]])
+				n_ext = int(np.min([n + self._n_seen_by_user[user], items_pool.shape[0]]))
 				rec = np.argpartition(allpreds, n_ext-1)[:n_ext]
 				seen = self.seen[self._st_ix_user[user] : self._st_ix_user[user] + self._n_seen_by_user[user]]
 				if self.reindex:
@@ -1352,6 +1407,9 @@ class CTPF:
 			Item(s) for which to predict for each user.
 		"""
 		assert self.is_fitted
+		cy = cy_float if self.use_float else cy_double
+		cython_loops = cython_loops_float if self.use_float else cython_loops_double
+
 		if isinstance(user, list) or isinstance(user, tuple):
 			user = np.array(user)
 		if isinstance(item, list) or isinstance(item, tuple):
@@ -1497,6 +1555,8 @@ class CTPF:
 		if new_max_id <= 0:
 			raise ValueError("Numeration of item IDs overlaps with IDs passed to '.fit'.")
 
+		cy = cy_float if self.use_float else cy_double
+		cython_loops = cython_loops_float if self.use_float else cython_loops_double
 		new_Theta_shp, temp = cy.calc_item_factors(
 					words_df, new_max_id, maxiter, cython_loops.cast_ind_type(self.k), stop_thr, random_seed, ncores,
 					cython_loops.cast_real_t(self.a), cython_loops.cast_real_t(self.b),
@@ -1591,6 +1651,8 @@ class CTPF:
 		## Will reuse the exact same function from the add items, need to change names accordingly
 		user_df.rename(columns={'UserId':'ItemId', 'AttributeId':'WordId'}, inplace=True)
 
+		cy = cy_float if self.use_float else cy_double
+		cython_loops = cython_loops_float if self.use_float else cython_loops_double
 		new_Omega_shp, temp = cy.calc_item_factors(
 					user_df, new_max_id, maxiter, cython_loops.cast_ind_type(self.k),
 					stop_thr, random_seed, ncores,
@@ -1622,6 +1684,8 @@ class CTPF:
 			msg = "Can only add " + err_subj + "s to a fitted model when called with 'keep_all_objs=True'."
 			raise ValueError(msg)
 
+		cython_loops = cython_loops_float if self.use_float else cython_loops_double
+
 		if ncores is None:
 			ncores = 1 
 		if ncores < 1:
@@ -1697,9 +1761,9 @@ class CTPF:
 		new_Theta = new_Theta_shp / self.Theta_rte
 		self.Theta_shp = np.r_[self.Theta_shp, new_Theta]
 		self.Theta = np.r_[self.Theta, new_Theta_shp]
-		self.Epsilon = np.r_[self.Epsilon, np.zeros((int(new_max_id), int(self.k)), dtype=ctypes.c_float)]
-		self.Epsilon_shp = np.r_[self.Epsilon_shp, np.zeros((int(new_max_id), int(self.k)), dtype=ctypes.c_float)]
-		self.Epsilon_rte = np.r_[self.Epsilon_rte, np.zeros((int(new_max_id), int(self.k)), dtype=ctypes.c_float)]
+		self.Epsilon = np.r_[self.Epsilon, np.zeros((int(new_max_id), int(self.k)), dtype=ctypes.c_float if self.use_float else ctypes.c_double)]
+		self.Epsilon_shp = np.r_[self.Epsilon_shp, np.zeros((int(new_max_id), int(self.k)), dtype=ctypes.c_float if self.use_float else ctypes.c_double)]
+		self.Epsilon_rte = np.r_[self.Epsilon_rte, np.zeros((int(new_max_id), int(self.k)), dtype=ctypes.c_float if self.use_float else ctypes.c_double)]
 		self._M2 = np.r_[self._M2, new_Theta]
 
 
@@ -1775,6 +1839,9 @@ class CTPF:
 		if (counts_df is None) and (not self._has_user_df):
 			raise ValueError("Must pass 'counts_df' to add a new user.")
 
+		cy = cy_float if self.use_float else cy_double
+		cython_loops = cython_loops_float if self.use_float else cython_loops_double
+
 		
 		if (counts_df is not None) and (user_df is not None) and self._has_user_df:
 			## factors based on both attributes and interactions
@@ -1826,8 +1893,8 @@ class CTPF:
 			self.Eta = np.r_[self.Eta, new_Eta]
 			self._M1 = np.r_[self._M1, new_Eta]
 			if self._has_user_df:
-				self.Omega = np.r_[self.Omega, np.zeros((new_max_id, self.k), dtype=ctypes.c_float)]
-				self.Omega_shp = np.r_[self.Omega_shp, np.zeros((new_max_id, self.k), dtype=ctypes.c_float)]
+				self.Omega = np.r_[self.Omega, np.zeros((new_max_id, self.k), dtype=ctypes.c_float if self.use_float else ctypes.c_double)]
+				self.Omega_shp = np.r_[self.Omega_shp, np.zeros((new_max_id, self.k), dtype=ctypes.c_float if self.use_float else ctypes.c_double)]
 
 		## factors based on user attributes
 		else:
@@ -1841,8 +1908,8 @@ class CTPF:
 			new_Omega = new_Omega_shp / self.Omega_rte
 			self.Omega_shp = np.r_[self.Omega_shp, new_Omega_shp]
 			self.Omega = np.r_[self.Omega, new_Omega]
-			self.Eta = np.r_[self.Eta, np.zeros((new_max_id, self.k), dtype=ctypes.c_float)]
-			self.Eta_shp = np.r_[self.Eta_shp, np.zeros((new_max_id, self.k), dtype=ctypes.c_float)]
+			self.Eta = np.r_[self.Eta, np.zeros((new_max_id, self.k), dtype=ctypes.c_float if self.use_float else ctypes.c_double)]
+			self.Eta_shp = np.r_[self.Eta_shp, np.zeros((new_max_id, self.k), dtype=ctypes.c_float if self.use_float else ctypes.c_double)]
 			self._M1 = np.r_[self._M1, new_Omega]
 
 		
@@ -1850,9 +1917,9 @@ class CTPF:
 		if self.keep_data and (counts_df is not None):
 			for u in range(int(new_max_id)):
 				items_this_user = counts_df.ItemId.values[counts_df.UserId == u]
-				self._n_seen_by_user = np.r_[self._n_seen_by_user, items_this_user.shape[0]]
-				self._st_ix_user = np.r_[self._st_ix_user, self.seen.shape[0]]
-				self.seen = np.r_[self.seen, items_this_user]
+				self._n_seen_by_user = np.r_[self._n_seen_by_user, items_this_user.shape[0]].astype(int)
+				self._st_ix_user = np.r_[self._st_ix_user, self.seen.shape[0]].astype(int)
+				self.seen = np.r_[self.seen, items_this_user].astype(int)
 		
 		## Adding the new IDs
 		if self.reindex:
@@ -1897,6 +1964,7 @@ class CTPF:
 		"""
 		assert self.is_fitted
 		HPF._process_valset(self, counts_df, valset=False)
+		cython_loops = cython_loops_float if self.use_float else cython_loops_double
 		out = {'llk': cython_loops.calc_llk(self.val_set.Count.values,
 											self.val_set.UserId.values,
 											self.val_set.ItemId.values,
